@@ -1,16 +1,18 @@
-import elbus
-import msgpack
 import os
 import sys
 import time
-from neotermcolor import colored
+import socket
 from functools import partial
+
+import elbus
+import msgpack
+from neotermcolor import colored
 from rapidtables import format_table, FORMAT_GENERATOR, FORMAT_GENERATOR_COLS
 
-from .tools import print_result, ok, edit_config, read_file, err
-from .tools import print_action_result, format_value, prepare_time
-from .tools import get_node_svc_info, edit_file, get_term_size
-from .client import call_rpc
+from .tools import print_result, ok, edit_config, read_file, write_file
+from .tools import print_action_result, format_value, prepare_time, err, warn
+from .tools import get_node_svc_info, edit_file, get_term_size, exec_cmd, xc
+from .client import call_rpc, DEFAULT_REPL_SERVICE
 from .sharedobj import common, current_command
 
 
@@ -21,6 +23,22 @@ def eva_control_c(c, pfx=''):
 
 class CLI:
 
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        for vc in [
+                'build', 'list', 'edit', 'config', 'add', 'remove', 'update'
+        ]:
+            setattr(self, f'venv_{vc}', partial(self.venv_cmd, vc))
+
+    def venv_cmd(self, vc, **kwargs):
+        args = vc
+        args += ' ' + ' '.join(kwargs.get('modules', []))
+        if kwargs.get('rebuild'):
+            args += ' --rebuild'
+        if kwargs.get('from_scratch'):
+            args += ' --from-scratch'
+        exec_cmd('venvmgr', args, search_in='sbin', search_system=False)
+
     def test(self):
         data = call_rpc('test')
         print_result(data, need_header=False, name_value=True)
@@ -61,6 +79,26 @@ class CLI:
     def server_status(self):
         eva_control_c('status')
 
+    def cloud_deploy(self, file, test, config_var):
+        args = (f'cloud deploy --timeout {current_command.timeout}'
+                f' --connection-path "{common.elbus_path}" "{file}"')
+        if config_var:
+            for c in config_var:
+                args += f' --config-var "{c}"'
+        if test:
+            args += ' --test'
+        exec_cmd('eva-cloud-manager', args)
+
+    def cloud_undeploy(self, file, test, config_var):
+        args = (f'cloud undeploy --timeout {current_command.timeout}'
+                f' --connection-path "{common.elbus_path}" "{file}"')
+        if config_var:
+            for c in config_var:
+                args += f' --config-var "{c}"'
+        if test:
+            args += ' --test'
+        exec_cmd('eva-cloud-manager', args)
+
     def update(self, repository_url, yes, info_only, test):
         import json
         from .tools import safe_download, set_file_lock
@@ -200,19 +238,34 @@ class CLI:
         finally:
             remove_file_lock('update')
 
-    def registry_manage(self, offline=False):
-        cmd = f'{common.dir_eva}/bin/eva-registry'
-        if offline:
-            cmd += ' --offline'
-        os.system(cmd)
+    def registry_manage(self, yedb_args=None):
+        cmd = f'{common.dir_eva}/venv/bin/yedb'
+        args = '' if yedb_args is None else ' ' + yedb_args
+        if not os.path.exists(cmd):
+            import shutil
+            cmd = shutil.which('yedb')
+            if not cmd:
+                raise RuntimeError('YEDB CLI not found')
+        try:
+            data = call_rpc('test')
+            system_name = data.get('system_name', '')
+            yedb_socket = f'elbus://{common.elbus_path}:eva.registry'
+            os.system(f'env "YEDB_PS=registry:{system_name}" "{cmd}" '
+                      f'"{yedb_socket}"{args}')
+        except:
+            import getch
+            warn('The node is offline, starting '
+                 'the registry manager in offline mode')
+            warn('The registry database will be '
+                 'LOCKED until the registry shell exit')
+            if yedb_args is None or getch.getch() != '\x03':
+                os.system(f'env YEDB_PS=registry "{cmd}"'
+                          f' "{common.dir_eva}/runtime/registry"{args}')
 
     def edit(self, fname, offline=False):
         if fname.startswith('config'):
-            cmd = f'{common.dir_eva}/bin/eva-registry'
-            if offline:
-                cmd += ' --offline'
-            cmd += f' edit eva/"{fname}"'
-            if not os.system(cmd) and fname != 'config/python-venv':
+            self.registry_manage(yedb_args=f'edit eva/"{fname}"')
+            if fname != 'config/python-venv':
                 print('restart or reload the node server to apply changes')
         else:
             edit_file(fname)
@@ -265,6 +318,217 @@ class CLI:
                     colored(row if full else row[:width],
                             **log_record_color(record['level'])))
 
+    def acl_list(self, acl_svc):
+        data = call_rpc('acl.list', target=acl_svc)
+        print_result(data, cols=['id', 'admin'])
+
+    def acl_edit(self, i, acl_svc):
+
+        def deploy(cfg, i):
+            call_rpc('acl.deploy', dict(acls=[cfg]), target=acl_svc)
+            print(f'ACL re-deployed: {i}')
+            print()
+
+        config = call_rpc('acl.get_config', dict(i=i), target=acl_svc)
+        edit_config(config, f'acl|{i}|config', deploy_fn=partial(deploy, i=i))
+
+    def acl_export(self, i, acl_svc, output=None):
+        import yaml
+        c = 0
+        configs = []
+        for acl in call_rpc('acl.list', target=acl_svc):
+            name = acl['id']
+            if (i.startswith('*') and name.endswith(i[1:])) or \
+                (i.endswith('*') and name.startswith(i[:-1])) or \
+                (i.startswith('*') and i.endswith('*') and i[1:-1] in name) or \
+                i == name:
+                c += 1
+                configs.append(acl)
+        dump = yaml.dump(dict(acls=configs), default_flow_style=False)
+        if output is None:
+            print(dump)
+        else:
+            write_file(output, dump)
+            print(f'{c} ACL(s) exported')
+            print()
+
+    def acl_deploy(self, acl_svc, file=None):
+        import yaml
+        acls = yaml.safe_load(read_file(file).decode()).pop('acls')
+        call_rpc('acl.deploy', dict(acls=acls), target=acl_svc)
+        print(f'{len(acls)} ACL(s) deployed')
+        print()
+
+    def acl_undeploy(self, acl_svc, file=None):
+        import yaml
+        acls = yaml.safe_load(read_file(file).decode()).pop('acls')
+        call_rpc('acl.undeploy', dict(acls=acls), target=acl_svc)
+        print(f'{len(acls)} ACL(s) undeployed')
+        print()
+
+    def acl_create(self, i, acl_svc):
+        call_rpc('acl.deploy', dict(acls=[dict(id=i)]), target=acl_svc)
+        ok()
+
+    def acl_destroy(self, i, acl_svc):
+        call_rpc('acl.destroy', dict(i=i), target=acl_svc)
+        ok()
+
+    def key_list(self, auth_svc):
+        data = call_rpc('key.list', target=auth_svc)
+        if not current_command.json:
+            for d in data:
+                d['acls'] = ', '.join(d['acls'])
+        print_result(data, cols=['id', 'acls'])
+
+    def key_edit(self, i, auth_svc):
+
+        def deploy(cfg, i):
+            call_rpc('key.deploy', dict(keys=[cfg]), target=auth_svc)
+            print(f'API key re-deployed: {i}')
+            print()
+
+        config = call_rpc('key.get_config', dict(i=i), target=auth_svc)
+        edit_config(config, f'key|{i}|config', deploy_fn=partial(deploy, i=i))
+
+    def key_export(self, i, auth_svc, output=None):
+        import yaml
+        c = 0
+        configs = []
+        for key in call_rpc('key.list', target=auth_svc):
+            name = key['id']
+            if (i.startswith('*') and name.endswith(i[1:])) or \
+                (i.endswith('*') and name.startswith(i[:-1])) or \
+                (i.startswith('*') and i.endswith('*') and i[1:-1] in name) or \
+                i == name:
+                c += 1
+                configs.append(key)
+        dump = yaml.dump(dict(keys=configs), default_flow_style=False)
+        if output is None:
+            print(dump)
+        else:
+            write_file(output, dump)
+            print(f'{c} API key(s) exported')
+            print()
+
+    def key_deploy(self, auth_svc, file=None):
+        import yaml
+        keys = yaml.safe_load(read_file(file).decode()).pop('keys')
+        call_rpc('key.deploy', dict(keys=keys), target=auth_svc)
+        print(f'{len(keys)} API key(s) deployed')
+        print()
+
+    def key_undeploy(self, auth_svc, file=None):
+        import yaml
+        keys = yaml.safe_load(read_file(file).decode()).pop('keys')
+        call_rpc('key.undeploy', dict(keys=keys), target=auth_svc)
+        print(f'{len(keys)} API key(s) undeployed')
+        print()
+
+    def key_create(self, i, auth_svc):
+        import random
+        import string
+        symbols = string.ascii_letters + '0123456789'
+        k = ''.join(random.choice(symbols) for i in range(16))
+        call_rpc('key.deploy', dict(keys=[dict(id=i, key=k)]), target=auth_svc)
+        data = call_rpc('key.get', dict(i=i), target=auth_svc)
+        print_result([data], cols=['id', 'key'])
+
+    def key_destroy(self, i, auth_svc):
+        call_rpc('key.destroy', dict(i=i), target=auth_svc)
+        ok()
+
+    def key_regenerate(self, i, auth_svc):
+        data = call_rpc('key.regenerate', dict(i=i), target=auth_svc)
+        print_result([data], cols=['id', 'key'])
+
+    def key_get(self, i, auth_svc):
+        data = call_rpc('key.get_config', dict(i=i), target=auth_svc)
+        if not current_command.json:
+            data['acls'] = ', '.join(data['acls'])
+        print_result([data], cols=['id', 'key', 'acls'])
+
+    def user_list(self, auth_svc):
+        data = call_rpc('user.list', target=auth_svc)
+        if not current_command.json:
+            for d in data:
+                d['acls'] = ', '.join(d['acls'])
+        print_result(data, cols=['login', 'acls'])
+
+    def user_edit(self, i, auth_svc):
+
+        def deploy(cfg, i):
+            call_rpc('user.deploy', dict(users=[cfg]), target=auth_svc)
+            print(f'user re-deployed: {i}')
+            print()
+
+        config = call_rpc('user.get_config', dict(i=i), target=auth_svc)
+        edit_config(config, f'user|{i}|config', deploy_fn=partial(deploy, i=i))
+
+    def user_export(self, i, auth_svc, output=None):
+        import yaml
+        c = 0
+        configs = []
+        for user in call_rpc('user.list', target=auth_svc):
+            name = user['login']
+            if (i.startswith('*') and name.endswith(i[1:])) or \
+                (i.endswith('*') and name.startswith(i[:-1])) or \
+                (i.startswith('*') and i.endswith('*') and i[1:-1] in name) or \
+                i == name:
+                c += 1
+                configs.append(user)
+        dump = yaml.dump(dict(users=configs), default_flow_style=False)
+        if output is None:
+            print(dump)
+        else:
+            write_file(output, dump)
+            print(f'{c} user(s) exported')
+            print()
+
+    def user_deploy(self, auth_svc, file=None):
+        import yaml
+        users = yaml.safe_load(read_file(file).decode()).pop('users')
+        call_rpc('user.deploy', dict(users=users), target=auth_svc)
+        print(f'{len(users)} user(s) deployed')
+        print()
+
+    def user_undeploy(self, auth_svc, file=None):
+        import yaml
+        users = yaml.safe_load(read_file(file).decode()).pop('users')
+        call_rpc('user.undeploy', dict(users=users), target=auth_svc)
+        print(f'{len(users)} user(s) undeployed')
+        print()
+
+    def user_create(self, i, auth_svc):
+        import pwinput
+        from hashlib import sha256
+        password = pwinput.pwinput()
+        call_rpc(
+            'user.deploy',
+            dict(users=[
+                dict(login=i, password=sha256(password.encode()).hexdigest())
+            ]),
+            target=auth_svc)
+        ok()
+
+    def user_destroy(self, i, auth_svc):
+        call_rpc('user.destroy', dict(i=i), target=auth_svc)
+        ok()
+
+    def user_password(self, i, auth_svc):
+        import pwinput
+        password = pwinput.pwinput()
+        data = call_rpc('user.set_password',
+                        dict(i=i, password=password),
+                        target=auth_svc)
+        ok()
+
+    def user_get(self, i, auth_svc):
+        data = call_rpc('user.get_config', dict(i=i), target=auth_svc)
+        if not current_command.json:
+            data['acls'] = ', '.join(data['acls'])
+        print_result([data], cols=['login', 'acls'])
+
     def svc_list(self):
         data = call_rpc('svc.list')
         print_result(data, cols=['id', 'status', 'pid', 'launcher'])
@@ -283,25 +547,24 @@ class CLI:
                 configs.append(
                     dict(id=name,
                          params=call_rpc('svc.get_params', dict(i=name))))
-        dump = yaml.dump(configs, default_flow_style=False)
+        dump = yaml.dump(dict(svcs=configs), default_flow_style=False)
         if output is None:
             print(dump)
         else:
-            with open(output, 'w') as fh:
-                fh.write(dump)
+            write_file(output, dump)
             print(f'{c} service(s) exported')
             print()
 
-    def svc_deploy(self, file=None, replace=False):
+    def svc_deploy(self, file=None):
         import yaml
-        svcs = yaml.safe_load(read_file(file).decode())
+        svcs = yaml.safe_load(read_file(file).decode()).pop('svcs')
         call_rpc('svc.deploy', dict(svcs=svcs))
         print(f'{len(svcs)} service(s) deployed')
         print()
 
     def svc_undeploy(self, file=None):
         import yaml
-        svcs = yaml.safe_load(read_file(file).decode())
+        svcs = yaml.safe_load(read_file(file).decode()).pop('svcs')
         call_rpc('svc.undeploy', dict(svcs=svcs))
         print(f'{len(svcs)} service(s) undeployed')
         print()
@@ -403,9 +666,10 @@ class CLI:
         params = dict(i=i, wait=wait)
         if priority is not None:
             params['priority'] = int(priority)
-        params['status'] = int(status)
+        action_params = {'status': int(status)}
         if value is not None:
-            params['value'] = format_value(value)
+            action_params['value'] = format_value(value)
+        params['params'] = action_params
         result = call_rpc('action', params)
         if current_command.json:
             import uuid
@@ -418,14 +682,16 @@ class CLI:
         params = dict(i=i, wait=wait)
         if priority is not None:
             params['priority'] = int(priority)
+        action_params = {}
         if arg:
-            params['args'] = [format_value(v) for v in arg]
+            action_params['args'] = [format_value(v) for v in arg]
         if kwarg:
             kw = {}
             for k in kwarg:
                 n, v = k.split('=', 1)
                 kw[n] = format_value(v)
-            params['kwargs'] = kw
+            action_params['kwargs'] = kw
+        params['params'] = action_params
         result = call_rpc('run', params)
         if current_command.json:
             import uuid
@@ -525,14 +791,96 @@ class CLI:
             print('total:', colored(data['items'], color='yellow'))
             print()
 
-    def node_list(self):
-        data = call_rpc('node.list')
-        for d in data:
-            if d['remote']:
-                d['type'] = 'remote'
-            else:
-                d['type'] = 'local'
-        print_result(data, cols=['name', 'svc', 'type', 'online'])
+    def node_list(self, repl_svc, default_rpl):
+        if default_rpl and repl_svc is None:
+            repl_svc = DEFAULT_REPL_SERVICE
+        if repl_svc is None:
+            data = call_rpc('node.list')
+            if not current_command.json:
+                for d in data:
+                    if d['remote']:
+                        d['type'] = 'remote'
+                    else:
+                        d['type'] = 'local'
+                    try:
+                        info = d.pop('info')
+                        d['version'] = info.get('version')
+                        d['build'] = info.get('build')
+                    except KeyError:
+                        pass
+            print_result(
+                data,
+                cols=['name', 'svc', 'type', 'online', 'version', 'build'])
+        else:
+            data = call_rpc('node.list', target=repl_svc)
+            print_result(data,
+                         cols=[
+                             'name', 'timeout', 'compress|n=compr',
+                             'ping_interval|n=ping int',
+                             'reload_interval|n=rel.int', 'static', 'enabled',
+                             'managed', 'online',
+                             'link_uptime|n=link upt|f=round:0', 'version',
+                             'build'
+                         ])
+
+    def node_append(self, i, repl_svc):
+        call_rpc('node.append', dict(i=i), target=repl_svc)
+        ok()
+
+    def node_edit(self, i, repl_svc):
+
+        def deploy_edited_node(cfg, i):
+            call_rpc('node.deploy', dict(nodes=[cfg]), target=repl_svc)
+            print(f'node re-deployed: {i}')
+            print()
+
+        config = call_rpc('node.get_config', dict(i=i), target=repl_svc)
+        edit_config(config,
+                    f'node|{i}|config',
+                    deploy_fn=partial(deploy_edited_node, i=i))
+
+    def node_export(self, i, repl_svc, output=None):
+        import yaml
+        configs = call_rpc('node.export', dict(i=i), target=repl_svc)
+        c = len(configs['nodes'])
+        dump = yaml.dump(configs, default_flow_style=False)
+        if output is None:
+            print(dump)
+        else:
+            write_file(output, dump)
+            print(f'{c} node(s) exported')
+            print()
+
+    def node_test(self, i, repl_svc):
+        call_rpc('node.test', dict(i=i), target=repl_svc)
+        ok()
+
+    def node_mtest(self, i, repl_svc):
+        call_rpc('node.mtest', dict(i=i), target=repl_svc)
+        ok()
+
+    def node_remove(self, i, repl_svc):
+        call_rpc('node.remove', dict(i=i), target=repl_svc)
+        ok()
+
+    def node_reload(self, i, repl_svc):
+        call_rpc('node.reload', dict(i=i), target=repl_svc)
+        print('reload request accepted')
+        print()
+
+    def node_deploy(self, repl_svc, file=None):
+        import yaml
+        nodes = yaml.safe_load(read_file(file).decode()).pop('nodes')
+        call_rpc('node.deploy', dict(nodes=nodes), target=repl_svc)
+        print(f'{len(nodes)} node(s) deployed')
+        print()
+
+    def node_undeploy(self, repl_svc, file=None):
+        import yaml
+        nodes = yaml.safe_load(read_file(file).decode()).pop('nodes')
+        call_rpc('node.undeploy', dict(nodes=nodes), target=repl_svc)
+        print(f'{len(nodes)} node(s) undeployed')
+        print()
 
     def item_list(self, i, n=None):
         data = call_rpc(
@@ -559,12 +907,11 @@ class CLI:
         for item in call_rpc('item.list', dict(i=i, node='.local')):
             c += 1
             configs.append(call_rpc('item.get_config', dict(i=item['oid'])))
-        dump = yaml.dump(configs, default_flow_style=False)
+        dump = yaml.dump(dict(items=configs), default_flow_style=False)
         if output is None:
             print(dump)
         else:
-            with open(output, 'w') as fh:
-                fh.write(dump)
+            write_file(output, dump)
             print(f'{c} item(s) exported')
             print()
 
@@ -589,16 +936,16 @@ class CLI:
         except KeyError:
             err(f'{i} has no status')
 
-    def item_deploy(self, file=None, replace=False):
+    def item_deploy(self, file=None):
         import yaml
-        items = yaml.safe_load(read_file(file).decode())
-        call_rpc('item.deploy', dict(items=items, replace=replace))
+        items = yaml.safe_load(read_file(file).decode()).pop('items')
+        call_rpc('item.deploy', dict(items=items))
         print(f'{len(items)} item(s) deployed')
         print()
 
     def item_undeploy(self, file=None):
         import yaml
-        items = yaml.safe_load(read_file(file).decode())
+        items = yaml.safe_load(read_file(file).decode()).pop('items')
         call_rpc('item.undeploy', dict(items=items))
         print(f'{len(items)} item(s) undeployed')
         print()
@@ -618,7 +965,7 @@ class CLI:
     def item_edit(self, i):
 
         def deploy_edited_item(cfg, i):
-            call_rpc('item.deploy', dict(items=[cfg], replace=True))
+            call_rpc('item.deploy', dict(items=[cfg]))
             print(f'item re-deployed: {i}')
             print()
 
@@ -818,6 +1165,58 @@ class CLI:
         val = call_rpc('lvar.decr', dict(i=i))
         print(val)
 
+    def system_update(self):
+        import shutil
+        apt = shutil.which('apt')
+        if apt:
+            update = 'apt update'
+            upgrade = 'apt -y upgrade'
+            clean = 'apt -y clean'
+        else:
+            apk = shutil.which('apk')
+            if apk:
+                update = 'apk update'
+                upgrade = 'apk upgrade'
+                clean = None
+            else:
+                raise RuntimeError(
+                    'Unsupported OS, please update the system manually')
+        if update:
+            xc(update,
+               ps='Updating package indexes',
+               verbose=current_command.debug)
+        if upgrade:
+            xc(upgrade, ps='Upgrading packages', verbose=current_command.debug)
+        if clean:
+            xc(clean, ps='Cleaning up', verbose=current_command.debug)
+        xc('sync', ps='Flushing data to disk', verbose=current_command.debug)
+        print('System updated')
+        print()
+
+    def system_reboot(self):
+        import getch
+        print('Reboot the system? (y/n) ', end='', flush=True)
+        c = getch.getch()
+        print()
+        if c in ['y', 'Y']:
+            warn('REBOOTING THE SYSTEM...')
+            if os.system('reboot'):
+                raise RuntimeError
+            while True:
+                time.sleep(1)
+
+    def system_poweroff(self):
+        import getch
+        print('Power off the system? (y/n) ', end='', flush=True)
+        c = getch.getch()
+        print()
+        if c in ['y', 'Y']:
+            warn('POWERING OFF...')
+            if os.system('poweroff'):
+                raise RuntimeError
+            while True:
+                time.sleep(1)
+
     def version(self):
         from . import __version__
         info = get_node_svc_info()
@@ -838,4 +1237,20 @@ class CLI:
                 'name': 'CLI version',
                 'value': __version__
             }]
+        try:
+            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
+            s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
+            s.connect(('255.255.255.255', 0))
+            ip = s.getsockname()[0]
+            try:
+                s.close()
+            except:
+                pass
+        except:
+            ip = None
+        if ip:
+            if current_command.json:
+                d['ip'] = ip
+            else:
+                d.append({'name': 'IP address', 'value': ip})
         print_result(d, need_header=False, name_value=False)
