@@ -7,6 +7,11 @@ import msgpack
 logger = logging.getLogger('elbus.client')
 logger.setLevel(logging.CRITICAL)
 
+DEFAULT_DB_SERVICE = 'eva.db.default'
+DEFAULT_REPL_SERVICE = 'eva.repl.default'
+DEFAULT_ACL_SERVICE = 'eva.aaa.acl'
+DEFAULT_AUTH_SERVICE = 'eva.aaa.localauth'
+
 
 def call_rpc(method, params=None, target='eva.core'):
     if current_command.debug:
