@@ -44,24 +44,20 @@ class ComplOIDtp:
 class ComplNode:
 
     def __call__(self, prefix, **kwargs):
-        for r in call_rpc('item.summary')['sources']:
-            yield r
+        for r in call_rpc('node.list'):
+            yield r['name']
 
 
 class ComplSvc:
 
-    def __call__(self, prefix, **kwargs):
-        for r in call_rpc('svc.list'):
-            yield r['id']
-
-
-class ComplDbSvc:
+    def __init__(self, tp=None):
+        self.prefix = None if tp is None else f'eva.{tp}.'
 
     def __call__(self, prefix, **kwargs):
         for r in call_rpc('svc.list'):
-            name = r['id']
-            if name.startswith('eva.db.'):
-                yield name
+            i = r['id']
+            if self.prefix is None or i.startswith(self.prefix):
+                yield i
 
 
 class ComplSvcRpcMethod:
@@ -93,20 +89,39 @@ class ComplYamlFile:
 
     def __call__(self, prefix, **kwargs):
         import glob
+        expanded = None
+
+        def expand_user(pfx):
+            nonlocal expanded
+            if pfx.startswith('~'):
+                expanded = os.path.expanduser('~')
+                return expanded + pfx[1:]
+            else:
+                return pfx
+
+        def contract_user(pfx):
+            nonlocal expanded
+            if expanded is None:
+                return pfx
+            else:
+                return '~' + pfx[len(expanded):]
+
         if not prefix:
             masks = ['*.yml', '*.yaml']
         elif prefix.endswith('.yml') or prefix.endswith('.yaml'):
             yield prefix
         elif prefix.endswith('.'):
+            prefix = expand_user(prefix)
             masks = [f'{prefix}*yml', f'{prefix}*yaml']
         else:
+            prefix = expand_user(prefix)
             masks = [f'{prefix}*.yml', f'{prefix}*.yaml']
         for mask in masks:
             for f in glob.glob(mask):
-                yield f
+                yield contract_user(f)
             for f in glob.glob(f'{prefix}*'):
                 if os.path.isdir(f):
-                    yield f'{f}/'
+                    yield contract_user(f'{f}/')
 
 
 class ComplEdit:
